/*

 *  (C) duber_ertw - ELE428 - Section N xxx-xxx-xxx

 *

 *  State Machine Simulation

 *  March 9, 2001

 *

*/



/* For the program specifications and known issues, please see the included

 * file "README", and for a graphical description of the state machine

 * being simulated, see the included file "diagram.ps"

*/



/* NOTE: Some might find my comments to be a little unorthodox, but we've got entertain the 

 * reader somehow. Either way, the end result is you should be able to understand everything,

 * and maybe even laugh a little....and by this point, you've probably marked alot of these labs,

 * and they're all starting to look the same. Enjoy. Here's a little variety.

*/



/* System Includes */

#include <stdio.h>

#include <"simState.h">



/* main() is repsonsible for handling the input and other interactions with other functions */

int main ()

{

  /* Declare and initialize variables as req's */

  int i, input, input_1, input_2;

  int garbage_flag = FALSE;

  int deletions_flag = FALSE;



  /* Define current_state variable and initialize to state 4 (E) */

  int current_state = E;



  /* Set inital states */

  set_states();

  init_garbage();



  /* Print initial state */

  fprintf(stdout, "\n\nThe current state is: E\n\n");



  while ((input = next_token()) != EOF)

    {

	 /* Go to next state as if a low signal was sent to input */

	 if (input == '0')

	   {

		if ((state[current_state].zero_path -> deleted) == FALSE)

		  {

		    current_state = (state[current_state].zero_path -> state_name) - 'A';

		    fprintf(stdout, "New Current State: %c\n", state[current_state].state_name);

		  }

	   }



	 /* Go to next state as if a high signal was sent to input */

	 else if (input == '1')

	   {

		if ((state[current_state].one_path -> deleted) == FALSE)

		  {

		    current_state = (state[current_state].one_path -> state_name) - 'A';

		    fprintf(stdout, "New Current State: %c\n", state[current_state].state_name);

		  }

	   }



	 /* Command to change the state machine */

	 else if (input == 'c')

	   {

		input_1 = next_token();

		input_2 = next_token();

  

		/* We can only change the link if the state we're trying to get to isn't deleted */

		if (state[(input_2 - 'A')].deleted == FALSE)

		  {

		    if (input_1 == '0')

			 {

			   state[current_state].zero_path = &state[input_2 - 'A'];

			 }

		    else if (input_1 == '1')

			 {

			   state[current_state].one_path = &state[input_2 - 'A'];

			 }

		    else /* Invalid input */

			 {

			   fprintf(stdout, "\nSorry, the syntax for that command is: c [0,1] [STATE]\n");

			 }

		  }

	   }



	 /* Print the state machine */

	 else if (input == 'p')

	   {

		print_states();

	   }



	 /* Delete command */

	 else if (input == 'd')

	   {

		input_1 = next_token();



		/* Do garbage collection first */

		init_garbage();

		find_garbage(current_state, 0);



		/* The user pressed enter right away, so assume no 2nd parameter.

		 * That means we've to deleted any garbage states we can find.

		 * NOTE: ASCII 10 is a carriage return.

		*/

		if (input_1 == 10) 

		  {

		    fprintf(stdout, "Deleted: ");

		    for (i = 0; i < 8; i++)

			 {

			    /* It's garbage, and not already deleted, delete it */

			   if ((garbage[i] == i) && (state[i].deleted == FALSE))

				{

				  state[i].deleted = TRUE;

				  deletions_flag = TRUE;

				  fprintf(stdout, "%c ", state[garbage[i]].state_name);

				}

			 }

		    /* Asthetic purposes */

		    fprintf(stdout, "\n");

			 

	 

		    /* No garbage to delete */

		    if (deletions_flag == FALSE)

			 fprintf(stdout, "No states deleted\n");

	

		    fprintf(stdout, "\n");



		  }

	   

		/* Assume we've been given a state to try to delete (input_1) */

		else  

		  {

		    /* It's already deleted or it's linked from somewhere else */

		    if ((state[(input_1 - 'A')].deleted == TRUE))

			 {

			   fprintf(stdout, "\nNot Deleted\n");

			 }

		    /* See if it's unreachable, if it is, and its not deleted then kill it */

		    else if ((is_reachable((input_1 - 'A')) == FALSE) && (state[(input_1 - 'A')].deleted == FALSE))

			 {

			   state[(input_1 - 'A')].deleted = TRUE;

			   fprintf(stdout, "\nDeleted\n");

			 }

		    else

		         {

			   fprintf(stdout, "\nNot Deleted\n");

			 }

		  }

	   }



	 /* Indentify garbage */

	 else if (input == 'g')

	   {

		init_garbage();

		find_garbage(current_state, 0);

	   

		/* Yeah, I know this is a shitty way to do things, but it's just a hack :-) */

		for (i = 0; i < 8; i++)

		  {

		    if (garbage[i] != -1 )

			 garbage_flag = TRUE;

		  }



		if (garbage_flag == FALSE)

		  {

		    fprintf(stdout, "\nNo Garbage\n");

		    fprintf(stdout, "\n");

		  }



		/* Garbage present */

		else

		  {

		    fprintf(stdout, "\nGarbage: ");

		    for (i = 0; i < 8; i++)

			 {

			   if (garbage[i] != -1)

				fprintf(stdout, "%c ", garbage[i] + 'A');

			 }

		    /* Make things look nicer */

		    fprintf(stdout, "\n");

		  }

	   }



	 else if (input == 'q' || input == 'Q')

	   {

		fprintf(stdout, "\n\nThanks for using Dave's State Machine Simulator...\n\n");

		return 0;

	   }



	 /* Invalid Input */

	 else

	   {

		/* Don't trip on an carriage return */

		if (input == 10);

		else

		  fprintf(stderr, "\nInvalid input. Valid commands are: [0,1,c,g,p,d]\n");

	   }

    }



  return 0;



}





/* Get input from stdin */

int next_token (void)

{

  int input;

  while (((input = getchar())==' ') && (input != '\n'));

  return input;

}





/* set_states(): creates initial state machine */

void set_states (void)

{

  state[A].state_name = 'A';

  state[A].zero_path = &state[F];

  state[A].one_path = &state[E];

  state[A].deleted = FALSE;



  state[B].state_name = 'B';

  state[B].zero_path = &state[H];

  state[B].one_path = &state[C];

  state[B].deleted = FALSE;



  state[C].state_name = 'C';

  state[C].zero_path = &state[A];

  state[C].one_path = &state[D];

  state[C].deleted = FALSE;



  state[D].state_name = 'D';

  state[D].zero_path = &state[B];

  state[D].one_path = &state[F];

  state[D].deleted = FALSE;



  state[E].state_name = 'E';

  state[E].zero_path = &state[C];

  state[E].one_path = &state[B];

  state[E].deleted = FALSE;



  state[F].state_name = 'F';

  state[F].zero_path = &state[D];

  state[F].one_path = &state[H];

  state[F].deleted = FALSE;



  state[G].state_name = 'G';

  state[G].zero_path = &state[E];

  state[G].one_path = &state[A];

  state[G].deleted = FALSE;



  state[H].state_name = 'H';

  state[H].zero_path = &state[G];

  state[H].one_path = &state[H];

  state[H].deleted = FALSE;

}



/* Prints listing of state machine to stdin */

void print_states(void)

{

  int i;



  /* A blank line for aesthetic purposes */

  fprintf(stdout, "\n");

  

  for (i=0; i < 8; i++)

    {  

	 if (!state[i].deleted)

	   {

		fprintf(stdout, "%c ", state[i].state_name);

		fprintf(stdout, "%c ", state[i].zero_path -> state_name);

		fprintf(stdout, "%c\n", state[i].one_path -> state_name);

	   }

    }



  /* Another blank */

  fprintf(stdout, "\n");



}



/* Finds all garbage states from the current state and store the names in garbage[8]

 * NOTE: This is a really ugly hack....needs some work before I'll like it.

   NOTE2: I've got a better way to do it below, this is just kept for the sake of sitting

   back and laughing at the effects of mixing beer and coding......

*/



void find_garbage1(int current_state)

{

  int i = 0;



  /* We can easily get stuck in an infinite loop here, so we've got be careful

   * How about setting a max number of recursion levels? Currently it is set to 16,

   *	as we can be a maximum of 8 jumps away from a state

  */



	 /* What we need to do is jump to the pointers of the current state,etc but as we go, see

	  *	if the current state is already listed in garbage[] - if it is, remove it.

	  *	- ie if we got here, it's not garbage.

	 */

  

	 /* Check if we're still listed in garbage */

	 for (i = 0; i < 8; i++)

	   {

		if (garbage[i] == current_state)

		  garbage[i] = -1;

	   }



	 /* increment the number of times we've recursed */

	 recursion_level++;



  if (recursion_level <= MAX_RECURSION_LEVEL/2)

    {

	 /* We can only go to a state if it's not deleted */

      /*if (!(state[current_state].zero_path -> deleted))*/

	   /*	   find_garbage(state[current_state].zero_path -> state_name - 'A');*/

    }

  if ((recursion_level > MAX_RECURSION_LEVEL/2) && (recursion_level <= MAX_RECURSION_LEVEL))

    {

      /*	 if (!(state[current_state].one_path -> deleted) == 0)

		 find_garbage(state[current_state].one_path -> state_name - 'A');*/

    }

}



/* The NEW and IMPROVED garbage collection function.....this one

 * actually works.  What we do is recurse along every hi and lo signal

 * path, as long as we are within MAX_RECURSION_LEVEL number of

 * recursion levels. This goes on the assumption that a state can

 * never be more than 8 jumps away.....right? Makes sense to me. We

 * pass in the current state, and the current level of recursion.

 */

void find_garbage(int current_state, int recursion_level)

{

  int i;



  for (i = 0; i < 8; i++)

  {

    if (garbage[i] == current_state)

      garbage[i] = -1;

  }



  if ((recursion_level < MAX_RECURSION_LEVEL) && !(state[current_state].zero_path -> deleted))

    {

	 recursion_level++;

	 find_garbage((state[current_state].zero_path -> state_name - 'A'), recursion_level);

	 recursion_level--;

    }





  if ((recursion_level < MAX_RECURSION_LEVEL) && !(state[current_state].one_path -> deleted))

    {

	 recursion_level++;

	 find_garbage((state[current_state].one_path -> state_name - 'A'), recursion_level);

	 recursion_level--;

    }



}



/* We need to clean up the garbage[] variable because it is dependant

 * on the state we're in.  Also need to reset recursion_level.

*/

void init_garbage(void)

{

  int i;



  recursion_level = 0;



  for (i = 0; i < 8; i++)

    garbage[i] = i;

}





/* Checks if a given state is reachable based on the garbage[] array,

   return 1 if so */

int is_reachable(int current_state)

{

  int i;



  for (i = 0; i < 8; i++)

    if (current_state == garbage[i])

	 return FALSE;



  return TRUE;



}



